import torch
from torch.nn.functional import mse_loss as mse
from .modules import DualSoftCriticModule
from .SoftActor import SoftActor
from ..utils import soft_update


class SoftCritic:
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
        lr: float,
        weight_decay: float,
        tau: float,
        gamma: float,
        alpha: float,
    ):
        self.tau: float = tau
        self.gamma: float = gamma
        self.alpha: float = alpha

        self.network = DualSoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )
        self.target = DualSoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )

        # copy source to target when initiated
        self.target.load_state_dict(self.network.state_dict())

        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=weight_decay
        )

    def train(
        self,
        states: torch.Tensor,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: SoftActor,
    ):
        self.network.train()
        self.optimizer.zero_grad()
        loss = self.loss(
            states=states,
            actions=actions,
            rewards=rewards,
            terminations=terminations,
            next_states=next_states,
            actor=actor,
        )
        loss.backward()
        self.optimizer.step()

    def loss(
        self,
        states: torch.Tensor,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: torch.nn.Module,
    ):
        q_alpha, q_beta = self.network(states, actions)
        target = self.calculate_target(
            rewards=rewards,
            terminations=terminations,
            next_states=next_states,
            actor=actor,
        )
        return mse(q_alpha, target) + mse(q_beta, target)

    @torch.no_grad()
    def calculate_target(
        self,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: SoftActor,
    ):
        termination_mask = 1 - terminations
        next_actions, next_probabilities = actor.sample(next_states)
        q_alpha, q_beta = self.target(next_states, next_actions)
        q = torch.min(q_alpha, q_beta)
        entropy_q = q - self.alpha * next_probabilities
        return rewards + self.gamma * termination_mask * entropy_q

    def update(self):
        soft_update(network=self.network.alpha, target=self.target.alpha, tau=self.tau)
        soft_update(network=self.network.beta, target=self.target.beta, tau=self.tau)



import torch


class SoftCriticModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
    ):
        super().__init__()

        self.state_dimension: int = state_dimension
        self.action_dimension: int = action_dimension
        self.hidden_dimension: int = hidden_dimension

        self.phi = torch.nn.Sequential(
            torch.nn.Linear(
                self.state_dimension + self.action_dimension, self.hidden_dimension
            ),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
        )
        self.fc = torch.nn.Linear(self.hidden_dimension, 1)

    def forward(self, state: torch.Tensor, action: torch.Tensor):
        psi = torch.cat([state, action], dim=1)
        features = self.phi(psi)
        return self.fc(features)



from .sac import SoftActor, SoftCritic
from .buffers import SoftActorCriticBuffer


class SoftActorCritic:
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int = 256,
        buffer_capacity: int = 1_000_000,
        batch_size: int = 256,
        actor_lr: float = 0.0002,
        critic_lr: float = 0.0003,
        actor_decay: float = 0.0,
        critic_decay: float = 0.0,
        tau: float = 0.005,
        alpha: float = 0.2,
        gamma: float = 0.99,
        min_std: float = -20,
        max_std: float = 2,
        warmup: int = 10_000,
    ):
        self.batch_size: int = batch_size
        self.warmup: int = warmup

        self.actor = SoftActor(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            lr=actor_lr,
            weight_decay=actor_decay,
            alpha=alpha,
            min_std=min_std,
            max_std=max_std,
        )
        self.critic = SoftCritic(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            lr=critic_lr,
            weight_decay=critic_decay,
            tau=tau,
            gamma=gamma,
            alpha=alpha,
        )
        self.buffer = SoftActorCriticBuffer(capacity=buffer_capacity)

    def train(self):
        # don't train before warmup
        if self.buffer.size < self.warmup:
            return

        data = self.buffer.sample_torch(self.batch_size)
        self.critic.train(
            states=data.states,
            actions=data.actions,
            rewards=data.rewards,
            terminations=data.terminations,
            next_states=data.next_states,
            actor=self.actor,
        )
        self.actor.train(states=data.states, critic_network=self.critic.network)
        self.critic.update()



import torch
from .modules import GraphAttentionModule
from ..core.concerns import MovesNetworkToDevice


class GraphAttention(MovesNetworkToDevice):
    def __init__(self, lr: float = 0.005, decay: float = 5e-4, dropout: float = 0.6):
        super().__init__()
        self.network = GraphAttentionModule()
        self.optimizer = torch.nn.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.criterion = torch.nn.CrossEntropyLoss()



import torch


class GraphAttentionHeadModule(torch.nn.Module):
    def __init__(self, in_features: int, out_features: int, alpha=0.2):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha

        self.W = torch.nn.Linear(in_features, out_features, bias=False)
        self.a_1 = torch.nn.Parameter(torch.randn(out_features, 1))
        self.a_2 = torch.nn.Parameter(torch.randn(out_features, 1))

        self.leaky_relu = torch.nn.LeakyReLU(self.alpha)
        self.elu = torch.nn.ELU()
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.W.weight, gain=1.414)
        torch.nn.init.xavier_uniform_(self.a_1, gain=1.414)
        torch.nn.init.xavier_uniform_(self.a_2, gain=1.414)

    def forward(self, node_features, adj_matrix):
        N = node_features.size(0)
        h_prime = self.W(node_features)
        s1 = torch.matmul(h_prime, self.a_1)
        s2 = torch.matmul(h_prime, self.a_2)
        e = s1 + s2.T
        e = self.leaky_relu(e)
        zero_vec = -9e15 * torch.ones_like(e)
        attention_mask = torch.where(
            adj_matrix > 0, e, zero_vec.to(node_features.device)
        )
        attention_weights = F.softmax(attention_mask, dim=1)
        h_new = torch.matmul(attention_weights, h_prime)
        return self.elu(h_new)



import torch
from .GraphAttentionHeadModule import GraphAttentionHeadModule


class MultiHeadGraphAttentionModule(torch.nn.Module):
    def __init__(self, num_heads: int, in_features: int, out_features: int, alpha=0.2):
        super().__init__()
        self.num_heads = num_heads
        self.head_out_features = out_features
        self.heads = torch.nn.ModuleList(
            [
                GraphAttentionHeadModule(in_features, out_features, alpha=alpha)
                for _ in range(num_heads)
            ]
        )

    def forward(
        self, node_features: torch.Tensor, adj_matrix: torch.Tensor
    ) -> torch.Tensor:
        outputs = [head(node_features, adj_matrix) for head in self.heads]
        h_new_concat = torch.cat(outputs, dim=1)
        return h_new_concat



from ...device import move


class MovesNetworkToDevice:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not hasattr(self, "network"):
            raise RuntimeError(
                "MovesNetworkToDevice assumes the underlying class has a network property."
            )

        move(self.network)
